/*
 *  Cube.cc
 *
 *  Cube class for Cube Polygonization Algorithm
 *
 *  Copyright (C) 2002 Tobias Dyckerhoff
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include "Cube.h"


list<Perm> Cube::rot; //list with the permutation group of the cube
vector<vector<double> > Cube::edge_vec(12);


/* Constructor initializing the permutations and in_count */
Cube::Cube(int i) : v(3), in(i), norm(3) {
  in_count = 0;
}

/* alternative Constructor setting up origin vector */
Cube::Cube(int i, double a, double b, double c) : v(3), in(i), norm(3) {
  in_count = 0;
  v[0] = a;
  v[1] = b;
  v[2] = c;
}


/* return the edge_vec index between i-th and j-th vertex */
unsigned char Cube::inbetw( unsigned char ii, unsigned char jj) {
  unsigned char i = (ii < jj)? ii : jj;
  unsigned char j = (ii < jj)? jj : ii;
  char edge_i = -1;
  

  /* The numbering of the edges is quite arbitrary, we chose this: */
  if( i == 0) {
    if( j == 1) 
      edge_i = 0;
    else if( j == 2)
      edge_i = 1;
    else if( j == 4)
      edge_i = 5;
  } 
  else if( i == 1) {
    if( j == 3) 
      edge_i = 3;
    else if( j == 5)
      edge_i = 4;
  }
  else if( i == 2) {
    if( j == 3) 
      edge_i = 2;
    else if( j == 6)
      edge_i = 6;
  }
  else if( i == 3) {
    if( j == 7) 
      edge_i = 7;
  }
  else if( i == 4) {
    if( j == 5)
      edge_i = 8;
    else if( j == 6)
      edge_i = 9;
  }
  else if( i == 5) {
    if( j == 7)
      edge_i = 11;
  }
  else if( i == 6) {
    if( j == 7)
      edge_i = 10;
  }
  
  if( edge_i == -1)
    cerr << "error: invalid edge pair" << endl;

  return edge_i;
}



/* uses permutations in tha cube to get the intersection 
   points and polygonize them */
list<unsigned char>* Cube::get_poly_index_list() {

  list<unsigned char>* index_list = new list<unsigned char>;
  if( in_count > 0 && in_count < 8) {
    int i = 0, j;
  
    /* get startin point inside of the surface */
    while( !in[i]){ i++;}

    /* try to move out of surface, using our permutation group */
    j = i;
    list<Perm>::iterator p = rot.begin();
    while( (i == j) && (p != rot.end()) ) {
      i = p->map(i);
      while( (i != j) && in[i]) 
      	i = p->map(i);
      if( i == j)
	p++;
    }

    /* ok, in this case all corners are IN but one. So we'll rotate
       i by one permutation. And do the same thing with map(i).
       This MUST hit the one corner which is OUT since 0 < in_count <8 */
    if ( p == rot.end()) {
      p = rot.begin();
      j = p->map(i);
      i = j;
      while( (i == j) && (p != rot.end()) ) {
	i = p->map(i);
	while( (i != j) && in[i]) 
	  i = p->map(i);
	if( i == j)
	  p++;
      }
      if ( p == rot.end()) {
	cerr << "derbo overflow" << endl; // this should not be possible
	exit(1);
      }
    }
      
    /* rotate to the next */
    for( ; !(in[p->map(i)]); i = p->map(i)) {};

    typedef list<unsigned char >::iterator l_it;
    l_it li;

    /* and iterate this until we are back to the first vertex */
    while( *li != inbetw(i, p->map(i)) ) {
      index_list->insert(index_list->end(), inbetw(i, p->map(i)));

      j = p->map(i);
      for( p = rot.begin(); p != rot.end() && i != p->map(j); p++){};
      i = j;
      
      for( i = p->map(i); ! in[p->map(i)]; i = p->map(i)) {};
      li = index_list->begin();
    }
  }
  
  return index_list;
}


/* returns the normal (it is invariant under translation -> used to cache it) */
vector<double>* Cube::get_normal(list<unsigned char >* poly_list) {

  vector<double> cur(3);
  vector<double> n(3);
  pol.clear();

  if( in_count > 0 && in_count < 8) {
    int i = 0, j;
    for(list<unsigned char >::iterator li = poly_list->begin();
	li != poly_list->end();
	li++) {
      for( int k = 0; k < 3; k++)
	cur[k] = v[k] + edge_vec[*li][k];     
      pol.insert(pol.end(), cur);
    }
        
    
    /* get normal vector of approximated "plane" */
    list<vector<double> >::iterator pv = pol.begin();
    vector<double> r = *pv;
    vector<double> w = *(++pv);
    vector<double> x(3);
    vector<double> y(3);

    n[0] = 1;
    n[1] = 1;
    n[2] = 1;
      
    
    if( pv != pol.end()) {
      if( ++pv != pol.end()) {
	for( int i = 0; i < 3; i++) {
	  x[i] = w[i] - r[i];
	  y[i] = (*pv)[i] - w[i];
	}
    
	n[0] = x[1]*y[2] - x[2]*y[1];
	n[1] = x[2]*y[0] - x[0]*y[2];
	n[2] = x[0]*y[1] - x[1]*y[0];
      
	double nabs = sqrt(n[0] * n[0]
			   +n[1] * n[1]
			   +n[2] * n[2]);
      
	for( int i = 0; i < 3; i++) {
	  n[i] /= nabs;
	}
      }
    } 

  }
  return new vector<double>(n);
}  



/* uses permutations in tha cube to get the intersection 
   points and polygonize them */
void Cube::polygonize(list<unsigned char >* poly_list, vector<double>* norm) {

  vector<double> cur(3);

  if( in_count > 0 && in_count < 8) {
    int i = 0, j;
    list<unsigned char >::iterator ni;
    for(list<unsigned char >::iterator li = poly_list->begin();
	li != poly_list->end();
	li++) {
      for( int k = 0; k < 3; k++)
	cur[k] = v[k] + edge_vec[*li][k];     
      pol.insert(pol.end(), cur);
    }

    /* save the given normal */
    this->norm = *norm;

  }
}


void Cube::init_rot() {
  
    /* (0264)o(1573) */
  unsigned int a[8] = { 2, 5, 6, 1, 0, 7, 4, 3};
  /* (0132)o(4675) */
  unsigned int b[8] = { 1, 3, 0, 2, 6, 4, 7, 5};
  /* (0451)o(2376) */
  unsigned int c[8] = { 4, 0, 3, 7, 5, 1, 2, 6};
  
  Cube::rot.push_front(Perm(8, a));
  Cube::rot.push_front(Perm(8, b));
  Cube::rot.push_front(Perm(8, c));
  
}


void Cube::init_edge_vec(double e) {
  double h = e / 2;
  
  vector<double> buf(3);

  buf[0] = 0;
  buf[1] = 0;
  buf[2] = h;
  edge_vec[0] = buf;

  buf[0] = 0;
  buf[1] = h;
  buf[2] = 0;
  edge_vec[1] = buf;

  buf[0] = 0;
  buf[1] = e;
  buf[2] = h;
  edge_vec[2] = buf;

  buf[0] = 0;
  buf[1] = h;
  buf[2] = e;
  edge_vec[3] = buf;

  buf[0] = h;
  buf[1] = 0;
  buf[2] = e;
  edge_vec[4] = buf;

  buf[0] = h;
  buf[1] = 0;
  buf[2] = 0;
  edge_vec[5] = buf;

  buf[0] = h;
  buf[1] = e;
  buf[2] = 0;
  edge_vec[6] = buf;

  buf[0] = h;
  buf[1] = e;
  buf[2] = e;
  edge_vec[7] = buf;

  buf[0] = e;
  buf[1] = 0;
  buf[2] = h;
  edge_vec[8] = buf;

  buf[0] = e;
  buf[1] = h;
  buf[2] = 0;
  edge_vec[9] = buf;

  buf[0] = e;
  buf[1] = e;
  buf[2] = h;
  edge_vec[10] = buf;

  buf[0] = e;
  buf[1] = h;
  buf[2] = e;
  edge_vec[11] = buf;

}
