/*
 *  QuatMath.h
 *
 *  Quaternion mathematics
 *
 *  Copyright (C) 2002 Tobias Dyckerhoff
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#ifndef QUAT_MATH_H
#define QUAT_MATH_H

#include <iostream.h>
#include <math.h>
#include "Vector3.h"



// TODO: memory handling is correct ????


/** Mathematical quaternion class. This class provides a mathematical
    interface to a Hamiltonian quaternion. */
class Quat {
  double vec[4];
  
 public:

  /** Constructor.*/
  Quat();
  /** Array constructor. Accepting the 4 doubles saved in an array.*/
  Quat(double *array);
  /** Element constructor. Accepting the 4 doubles given separately.*/
  Quat(double a, double b, double c, double d);
  

  /** Equal. this == q ?*/
  bool operator==(const Quat &q) const;
  /** Unequal. this != q ?*/
  bool operator!=(const Quat &q) const;
  /** Assignment. this = q */
  Quat& Quat::operator=(const Quat &q);
  /** Multiplication. this * q */
  Quat& operator*(const Quat &q) const;
  /** Addition. this + q */
  Quat& operator+(const Quat &q) const;
  /** Subtraction. this - q */
  Quat& operator-(const Quat &q) const;
  /** Multiplication=. this *= q */
  Quat& operator*=(const Quat &q);
  /** Addition=. this += q */
  Quat& operator+=(const Quat &q);
  /** Inversion. (this)^(-1)*/
  Quat& inv() const;
  /** Norm. ||this||*/
  double abs() const;
  /** Get the coordinate vector. */
  double* get_vec() const;
  /** Quadratic iteration function. (this)^2 + c */
  Quat& f(const Quat &c);
  /** Cubic iteration function. (this)^3 + c */
  Quat& g(const Quat &c);

};


/** Quaternion rotation. Rotates v around axis by angle. 
*/
inline Vector3 q_rotate(Vector3 &v, double angle, Vector3 &axis){
  Quat qv(0,v[0],v[1],v[2]);
  double s=sin(angle);
  Quat rot(cos(angle), 
	   s*axis[0],
	   s*axis[1],
	   s*axis[2]);  
  Quat rot_=rot.inv();
  qv=rot*qv*rot_;
  
  double* qv_vec=qv.get_vec();
  Vector3 res(qv_vec[1],qv_vec[2],qv_vec[3]);
  return res;
}



inline Quat& Quat::operator+=(const Quat &q) {
  Quat *res = new Quat();

  for( int i = 0; i < 4; i++) 
    res->vec[i] = vec[i] + q.vec[i];

  *this = *res;
  delete res;
  return *this;
}

inline Quat& Quat::operator*=(const Quat &q) {
  Quat *res = new Quat();
  res->vec[0] = vec[0]*q.vec[0] 
    - vec[1]*q.vec[1] 
    - vec[2]*q.vec[2]
    - vec[3]*q.vec[3];

  res->vec[1] = vec[0]*q.vec[1]
    + vec[1]*q.vec[0]
    + vec[2]*q.vec[3]
    - vec[3]*q.vec[2];

  res->vec[2] = vec[0]*q.vec[2]
    + vec[2]*q.vec[0]
    + vec[3]*q.vec[1]
    - vec[1]*q.vec[3];

  res->vec[3] = vec[0]*q.vec[3]
    + vec[3]*q.vec[0]
    + vec[1]*q.vec[2]
    - vec[2]*q.vec[1];
  
  *this = *res;

  delete res;
  return *this;
}



inline Quat& Quat::f(const Quat &c) {
  *this *= *this;
  *this += c;
  
  return *this;
}


inline Quat& Quat::g(const Quat &c) {
  *this *= *this * *this;
  *this += c;
  
  return *this;
}


/** Standard c++ output for quaternions. Use cout << q.*/
ostream& operator<<(ostream &s, const Quat &q);

#endif QUAT_MATH_H



