/*
 *  Gtk_PathMaster.h
 *
 *  Master of waypoint path
 *
 *  Copyright (C) 2002 Tobias Dyckerhoff
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#ifndef waypoint_H
#define waypoint_H

#include <vector>
#include "../SpaceEntity.h"
#include <gtk--/window.h>
#include "../Vector3.h"
#include "Gtk_FloatEntry.h"
#include "Gtk_CoordEntry.h"

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include <gtk--/buttonbox.h>
#include <gtk--/entry.h>
#include <gtk--/table.h>
#include <gtk--/frame.h>
#include <gtk--/button.h>
#include <gtk--/box.h>

/** Desribes a wapoint on the camera flight path. Is derived from SpaceEntity,
    so it has position and orientation information, and additionaly some 
    waypoint specific information like time and a forth coordinate. */

class Waypoint: public SpaceEntity
{
 public:
  /** Constructor. */
  Waypoint();

  /** Time at which the Camera should arrive during its flight. */
  double time; 

  /** Our quaternion space is 4D, so the waypoint has an additional coordinate. */
  double x4; 
  
  /** Set the time. */
  int set_time(double p);

  /** Set the forth coordinate.*/
  int set_x4(double p); 

  /** Signal which is emitted when x4 has changed. */  
  SigC::Signal1<int, double>x4_update;

  /** Signal which is emitted when the time has changed. */  
  SigC::Signal1<int, double>time_update;

};


/** This is the window through which the camera flight path can be edited. All 
    handling is done by the PathMaster.Currently this window is actually destroyed
    when closed and newly constructed when opened, which is not the case with most
    other windows (like CameraControl). This was done to save memory, but it might
    be changed to the common hide&show way because it seems to be problematic to
    set the window transient now (so that it always keeps being in front of the
    QuatFractViewer). */
class WaypointEditor: public Gtk::Window
{

  /* all is public so the PathMaster can overtake control.*/
 public:
 Gtk::Table table;
 Gtk::Frame frame;
 Gtk::Frame wpframe;
 Gtk::Button get_cam_b;
 Gtk::Button insert_b;
 Gtk::Button replace_b;
 Gtk::Button erase_b;
 Gtk::Button next_wp_b;
 Gtk::Button prev_wp_b;
 Gtk::VBox vbbox;
 Gtk::VBox vwpbox;
 Gtk::Entry wp_entry;
 CoordEntry pos;
 CoordEntry lookat;
 CoordEntry up;
 FloatEntry x4;
 FloatEntry time;

 /** Set the entries with the stats of wp.*/
 void update_coord_entries(const Waypoint& wp);

 /** Constructor */
 WaypointEditor();
};

/** This class manages the camera flight path. Currently the editor window is 
    completly handled by this class. Might be changed in the future (see 
    WaypointEditor class for details). 
    The PathMaster contains the path of waypoints.*/

class PathMaster 
{
  /** Here all the waypoints are stored. */
  vector<Waypoint*> path; 

  /** Tells status of the PathMaster/ the editor window. */
  int active; 

  /** A new Waypoint (defined through the entries of the Editor window, stored in
      new_wp) is inserted into the path after the one defind by "current". This 
      method is connected to the corresponding button of the WapoinEditor window. */
  void add_waypoint(); 

  /** The "current" Waypoint is deleted. This method is connected to the 
      corresponding button of the WapoinEditor window. */
  void delete_waypoint(); 

  /** The "current" Waypoint is replaced by a new one with the stats defined 
      through the entries in the Editor window (stored in new_wp). This method is
      connected to the corresponding button of the WapoinEditor window. */
  void replace_waypoint();

  /** The next Waypoint within the path is selected and set current.*/
  void next_waypoint();

  /** The previous Waypoint within the path is selected and set current.*/
  void prev_waypoint();
  
  /** Defines the current Waypoint choosed within the editor. */
  int  current; 
  
  /** Here all the stats which are currently in the entries of the editor are
      stored. */
  Waypoint new_wp; 

 public:
  /** Constructor.*/
  PathMaster();  
  
  /** The WaypointEditor window. */
  WaypointEditor* editor;

  /** This signal is supposed to recieve the current stats of the camera, which
      are needed for the "get stats from camera" button. */
  SigC::Signal0<SpaceStats> camera_stats_input; 

  /** When the stats of a Waypoint have changed, the scene needs to be redrawn. */
  SigC::Signal0<int> redraw; 

 
  //   SigC::Signal0<void> toggle_my_tool;
  
  /** This signal passes waypoint changes to the Gtk_PovrayOut (connection is 
      done by the QuatFractViewer).*/
  SigC::Signal0<void> path_update;

  /** The stats recieved from path_update signal are stored within new_wp. */
  void get_camera_pressed();

  /** One of the entries in the editor was activated. Store valid information
      within new_wp.*/
  void entry_activated(int id);

  //Ignore this
  void register_data (SigC::ScopeNode *);
  
  /** Is called when the PathMaster button in the toolbar was pressed. Sets active
     to 1 and instantiates a new WaypointEditor window, doing all the initializing
     like connecting the editors signals. */
  void activate(); 

  /** The WapointEditor window is destroyed. */ 
  void deactivate(); 

  /** Here the wapoints are drawn. Is called within the draw_gl_context_imp
      of the QuatFractViewer.*/
  void draw(); 
  
  /** Returns the path. */
  vector< Waypoint*> get_path() { return path;}

};







#endif
